︠1c572d9d-58f2-463f-9127-ba1260739769i︠
%md

### Playing with Finite Fields...

Here we go over some of the commands you can use to explore finite fields using SAGE.

The SAGE manual is [here](http://doc.sagemath.org/html/en/reference/finite_rings/sage/rings/finite_rings/finite_field_constructor.html)

The [original Rijndael paper](https://csrc.nist.gov/csrc/media/projects/cryptographic-standards-and-guidelines/documents/aes-development/rijndael-ammended.pdf) is very helpful.
︡e10814bd-04a2-4d18-957e-ef9a0e602168︡{"done":true,"md":"\n### Playing with Finite Fields...\n\nHere we go over some of the commands you can use to explore finite fields using SAGE.\n\nThe SAGE manual is [here](http://doc.sagemath.org/html/en/reference/finite_rings/sage/rings/finite_rings/finite_field_constructor.html)\n\nThe [original Rijndael paper](https://csrc.nist.gov/csrc/media/projects/cryptographic-standards-and-guidelines/documents/aes-development/rijndael-ammended.pdf) is very helpful."}
︠1f506789-aab5-47e7-ab4b-529dfcc58cb2i︠
%md

### Galois Fields

In this worksheet we will explore finite fields, also sometimes called Galois Fields after

[Evariste Galois](https://en.wikipedia.org/wiki/%C3%89variste_Galois)

As we saw in a previous slide deck, there is a unique finite field of order $p^k$ for each prime $p$ and positive integer $k$.

Recall also that to be a field $(F,\times,+)$ means that both $(F-\{0\},\times)$ and $(F,+)$ are groups.

The hard condition to satisfy is usually multiplicative inverses.

$GF(p^k)$ denotes the unique finite field of order $p^k$.

The elements of this group are polynomials with coefficients mod $p$ of degree at most $k-1$.

Multiplication is modulo a "prime" polynomial, which stops degree from increasing when two polynomials are multiplied.

We really only care about the case when $p=2$ and $k=8$.

But $GF(2^8)$ is kind of big, so we will warm up with $GF(2^4)$.




︡91b80b7d-219a-45f0-86ef-28639476fd54︡{"done":true,"md":"\n### Galois Fields\n\nIn this worksheet we will explore finite fields, also sometimes called Galois Fields after\n\n[Evariste Galois](https://en.wikipedia.org/wiki/%C3%89variste_Galois)\n\nAs we saw in a previous slide deck, there is a unique finite field of order $p^k$ for each prime $p$ and positive integer $k$. \n\nRecall also that to be a field $(F,\\times,+)$ means that both $(F-\\{0\\},\\times)$ and $(F,+)$ are groups.\n\nThe hard condition to satisfy is usually multiplicative inverses.\n\n$GF(p^k)$ denotes the unique finite field of order $p^k$.\n\nThe elements of this group are polynomials with coefficients mod $p$ of degree at most $k-1$.\n\nMultiplication is modulo a \"prime\" polynomial, which stops degree from increasing when two polynomials are multiplied.\n\nWe really only care about the case when $p=2$ and $k=8$.\n\nBut $GF(2^8)$ is kind of big, so we will warm up with $GF(2^4)$."}
︠14b39d11-3aab-4753-b71f-5162f85ed971s︠
## Define the field.  Fields are traditionally denoted k.
k = GF(2^4,'x')
︡0b6068ee-1bf9-47a8-a956-b31217b980e3︡{"done":true}
︠4f9922d9-79e3-42fe-9d9b-138691a18db2s︠
## Tell the object to print itself...
k
︡3f697491-6d08-4baf-8109-4c52ea6d4a4f︡{"stdout":"Finite Field in x of size 2^4\n"}︡{"done":true}
︠d65c61b7-e707-43aa-8f60-c643883d213bs︠
## Have a look at all the elements...
for i,p in enumerate(k):
    print(i,p)
︡31678da4-9391-4957-bc8b-5127ccb76165︡{"stdout":"0 0\n1 x\n2 x^2\n3 x^3\n4 x + 1\n5 x^2 + x\n6 x^3 + x^2\n7 x^3 + x + 1\n8 x^2 + 1\n9 x^3 + x\n10 x^2 + x + 1\n11 x^3 + x^2 + x\n12 x^3 + x^2 + x + 1\n13 x^3 + x^2 + 1\n14 x^3 + 1\n15 1\n"}︡{"done":true}
︠b58e8438-5890-4d93-97fc-4bcf1884e24fs︠
k.modulus()
︡e69e221c-45e7-4ad6-ad2b-ae484e2f0aa8︡{"stdout":"x^4 + x + 1\n"}︡{"done":true}
︠2d609e72-85c0-4a24-895a-56a6508b743bs︠
## Addition example...
︡3a071ced-d157-4dc1-9954-cbf34a4d8594︡{"done":true}
︠08a45724-dc2e-4892-be9d-5b63d45176bds︠
k("x^2")+k("x^3+x+1")
︡c798358e-0e17-486c-8827-b41df2be5898︡{"stdout":"x^3 + x^2 + x + 1\n"}︡{"done":true}
︠2941f730-0aa7-4847-93eb-b36e3871b3c3s︠
### Multiplication example
︡fd44c034-2b4b-4ad0-ac5c-d4f98061dba7︡{"done":true}
︠97013a19-dbf8-4ce9-8834-a46ccf703d7es︠
p = k("x^2+x+1")
q = k("x^2+1")
p*q

︡14207094-ecaf-4419-ac40-b2cd7355d954︡{"stdout":"x^3\n"}︡{"done":true}
︠52fec5ca-b315-40e7-a77f-77084580fcb0s︠
1/p
︡5ee7b52c-e886-4ea3-ad9a-6df080933729︡{"stdout":"x^2 + x\n"}︡{"done":true}
︠0ab68ae7-fb41-4790-97e1-99df7ee0c2a3s︠
p*1/p
︡1e9c736f-9277-4c29-83fe-909d21c1e763︡{"stdout":"1\n"}︡{"done":true}
︠4d5b0e24-cb79-441c-b1f4-c3be326b147a︠
︡96071a12-2062-4ddb-a664-f5f5b75a3f54︡
︠5e7bea88-ef25-45cd-906c-c600dfae7e45s︠
## Get a list of all the elements...
elements = list(k)
elements
︡0e334f11-36f0-411d-9fba-a580efd35cad︡{"stdout":"[0, x, x^2, x^3, x + 1, x^2 + x, x^3 + x^2, x^3 + x + 1, x^2 + 1, x^3 + x, x^2 + x + 1, x^3 + x^2 + x, x^3 + x^2 + x + 1, x^3 + x^2 + 1, x^3 + 1, 1]\n"}︡{"done":true}
︠c9a4bcf6-2f47-4e8b-8ab0-629ff72465ecs︠
## Pick two of the elements to examine...
A = elements[7]
B = elements[10]
A,B
︡4a876edf-e31c-4200-9df4-42c1358a30c8︡{"stdout":"(x^3 + x + 1, x^2 + x + 1)\n"}︡{"done":true}
︠3ad97058-87cd-4906-98ce-ac8d7c1f9bcfs︠
A+B
︡4231c1ef-a8a4-4af9-bbf1-d3ba2109d0d1︡{"stdout":"x^3 + x^2\n"}︡{"done":true}
︠4f674f8b-3255-4e0e-97ae-e8c37926692bs︠
## Modulo 2 addition and subtraction are the same thing...
A-B
︡c6be7dbd-1b36-460b-87a1-c5f618b9d0f6︡{"stdout":"x^3 + x^2\n"}︡{"done":true}
︠c382058d-b54b-4bc6-9eec-6df1f6035c3as︠
A*B
︡37e5c1ba-e118-4965-a8d1-32f5925f6662︡{"stdout":"x^2\n"}︡{"done":true}
︠0d748620-f8e2-4d28-9bc0-fcd579a512f3s︠
A_inverse = 1/A ### This is metaphorical!
A_inverse
︡176eed51-ec20-4e31-a40b-534cfc4ec0ec︡{"stdout":"x^2 + 1\n"}︡{"done":true}
︠052c2304-3871-42fd-94b0-40aebd2e572bs︠
A*A_inverse
︡435441e7-8a51-49b7-8f20-00e1dd260cf6︡{"stdout":"1\n"}︡{"done":true}
︠10adf77c-a1f7-41c4-9a57-cf5bed099649s︠
A.multiplicative_order()
︡8cc89d23-c2ef-4c58-b293-f02749b0d388︡{"stdout":"15\n"}︡{"done":true}
︠0fcd9f62-a6eb-4ceb-9614-73902e8d0d62s︠
powers_of_A = set([A**i for i in range(16)])
print(len(powers_of_A))
powers_of_A
︡be0c0bd5-3b89-46b7-9fcf-2baac4e2a889︡{"stdout":"15\n"}︡{"stdout":"{1, x, x + 1, x^2, x^2 + 1, x^2 + x, x^2 + x + 1, x^3, x^3 + 1, x^3 + x, x^3 + x + 1, x^3 + x^2, x^3 + x^2 + 1, x^3 + x^2 + x, x^3 + x^2 + x + 1}\n"}︡{"done":true}
︠a3105679-9a29-4d05-a1f0-d47d6cbce8a1s︠
k.is_field()
︡634e9dda-c364-46ed-a0b8-dcb6e8f1c46c︡{"stdout":"True\n"}︡{"done":true}
︠b724e1bd-8af5-42b0-bc8f-bf449e64b12bs︠
k.is_finite()
︡52d98eb4-ab1a-4c06-8f4f-bfbfc80cc1c4︡{"stdout":"True\n"}︡{"done":true}
︠0a9e8d13-c9e0-47b3-a1d0-7c28ca4316c1s︠
k.addition_table()
︡df71c2c7-59d5-4acf-8878-ee5d35c1a8e2︡{"stdout":"+  a b c d e f g h i j k l m n o p\n +--------------------------------\na| a b c d e f g h i j k l m n o p\nb| b a d c f e h g j i l k n m p o\nc| c d a b g h e f k l i j o p m n\nd| d c b a h g f e l k j i p o n m\ne| e f g h a b c d m n o p i j k l\nf| f e h g b a d c n m p o j i l k\ng| g h e f c d a b o p m n k l i j\nh| h g f e d c b a p o n m l k j i\ni| i j k l m n o p a b c d e f g h\nj| j i l k n m p o b a d c f e h g\nk| k l i j o p m n c d a b g h e f\nl| l k j i p o n m d c b a h g f e\nm| m n o p i j k l e f g h a b c d\nn| n m p o j i l k f e h g b a d c\no| o p m n k l i j g h e f c d a b\np| p o n m l k j i h g f e d c b a\n"}︡{"stdout":"\n"}︡{"done":true}
︠7d718afc-89fc-49e0-8d9e-4acd437f22fcs︠
k.multiplication_table()
︡a6a27771-30a1-44e0-af20-d32f49541a6f︡{"stdout":"*  a b c d e f g h i j k l m n o p\n +--------------------------------\na| a a a a a a a a a a a a a a a a\nb| a b c d e f g h i j k l m n o p\nc| a c e g i k m o d b h f l j p n\nd| a d g f m p k j l i n o h e b c\ne| a e i m d h l p g c o k f b n j\nf| a f k p h c n i o l e b j m d g\ng| a g m k l n h b f d j p o i c e\nh| a h o j p i b g n k d e c f m l\ni| a i d l g o f n m e p h k c j b\nj| a j b i c l d k e n f m g p h o\nk| a k h n o e j d p f i c b l g m\nl| a l f o k b p e h m c j n g i d\nm| a m l h f j o c k g b n p d e i\nn| a n j e b m i f c p l g d o k h\no| a o p b n d c m j h g i e k l f\np| a p n c j g e l b o m d i h f k\n\n"}︡{"done":true}
︠335e752c-be1b-497d-95b4-64ddcbc518b1s︠
## The irreducible polynomial for k...
k.modulus()
︡c56648fc-1801-448a-8825-e0157e7d5e04︡{"stdout":"x^4 + x + 1\n"}︡{"done":true}
︠5d509c0d-840d-4cc4-a000-a1c1e44a6902i︠
%md

### What about GF($2^8$)?
The actual field used in AES is GF($2^8$).
Let's look at that field...

︡99479fab-43dc-4702-bc1a-a1fac59b7ad9︡{"done":true,"md":"\n### What about GF($2^8$)?\nThe actual field used in AES is GF($2^8$).\nLet's look at that field..."}
︠99df5bab-3434-47ed-8c62-6cf8e005b2a4s︠
R = PolynomialRing(GF(2),'x')
p=R("x^8+x^4+x^3+x+1")
K = GF(2**8,'x',modulus=p)
K.modulus()
︡260fece0-7056-4cf9-94a5-ba9dfea1ea23︡{"stdout":"x^8 + x^4 + x^3 + x + 1\n"}︡{"done":true}
︠c24aa99c-14fb-4b90-9169-c6f0c6f943abs︠
elements = K.list()
elements
︡be6a34b4-cb38-4292-b2cf-0b2f20f40e2e︡{"stdout":"[0, x + 1, x^2 + 1, x^3 + x^2 + x + 1, x^4 + 1, x^5 + x^4 + x + 1, x^6 + x^4 + x^2 + 1, x^7 + x^6 + x^5 + x^4 + x^3 + x^2 + x + 1, x^4 + x^3 + x, x^5 + x^3 + x^2 + x, x^6 + x^5 + x^4 + x, x^7 + x^4 + x^2 + x, x^7 + x^5 + 1, x^7 + x^6 + x^5 + x^4 + x^3, x^4 + x + 1, x^5 + x^4 + x^2 + 1, x^6 + x^4 + x^3 + x^2 + x + 1, x^7 + x^6 + x^5 + 1, x^5 + x^4 + x^3, x^6 + x^3, x^7 + x^6 + x^4 + x^3, x^6 + x^5 + x^4 + x + 1, x^7 + x^4 + x^2 + 1, x^7 + x^5 + x^2, x^7 + x^6 + x^5 + x^4 + x^2 + x + 1, x, x^2 + x, x^3 + x, x^4 + x^3 + x^2 + x, x^5 + x, x^6 + x^5 + x^2 + x, x^7 + x^5 + x^3 + x, x^7 + x^6 + x^5 + x^2 + 1, x^5 + x^4 + x^2, x^6 + x^4 + x^3 + x^2, x^7 + x^6 + x^5 + x^2, x^5 + x^4 + x^2 + x + 1, x^6 + x^4 + x^3 + 1, x^7 + x^6 + x^5 + x^3 + x + 1, x^5 + x^2 + x, x^6 + x^5 + x^3 + x, x^7 + x^5 + x^4 + x^3 + x^2 + x, x^7 + x^6 + x^4 + x^3 + 1, x^6 + x^5 + x^4, x^7 + x^4, x^7 + x^5 + x^3 + x + 1, x^7 + x^6 + x^5 + x^2 + x, x^5 + x^4 + 1, x^6 + x^4 + x + 1, x^7 + x^6 + x^5 + x^4 + x^2 + 1, x^2, x^3 + x^2, x^4 + x^2, x^5 + x^4 + x^3 + x^2, x^6 + x^2, x^7 + x^6 + x^3 + x^2, x^6 + x^3 + x^2 + x + 1, x^7 + x^6 + x^4 + 1, x^6 + x^5 + x^3, x^7 + x^5 + x^4 + x^3, x^7 + x^6 + x^4 + x + 1, x^6 + x^5 + x^3 + x^2 + x, x^7 + x^5 + x^4 + x, x^7 + x^6 + x^3 + x^2 + 1, x^6 + x^3 + x^2, x^7 + x^6 + x^4 + x^2, x^6 + x^5 + x^2 + x + 1, x^7 + x^5 + x^3 + 1, x^7 + x^6 + x^5, x^5 + x^4 + x^3 + x + 1, x^6 + x^3 + x^2 + 1, x^7 + x^6 + x^4 + x^2 + x + 1, x^6 + x^5 + x, x^7 + x^5 + x^2 + x, x^7 + x^6 + x^5 + x^4 + 1, x^3, x^4 + x^3, x^5 + x^3, x^6 + x^5 + x^4 + x^3, x^7 + x^3, x^7 + x + 1, x^7 + x^4 + x^3 + x^2 + x, x^7 + x^5 + x^4 + x^3 + 1, x^7 + x^6 + x^4, x^6 + x^5 + x^3 + x + 1, x^7 + x^5 + x^4 + x^3 + x^2 + 1, x^7 + x^6 + x^4 + x^3 + x^2, x^6 + x^5 + x^4 + x^3 + x^2 + x + 1, x^7 + 1, x^7 + x^4 + x^3, x^7 + x^5 + x^4 + x + 1, x^7 + x^6 + x^3 + x^2 + x, x^6 + x^3 + 1, x^7 + x^6 + x^4 + x^3 + x + 1, x^6 + x^5 + x^4 + x^2 + x, x^7 + x^4 + x^3 + x, x^7 + x^5 + x^4 + x^2 + 1, x^7 + x^6 + x^2, x^6 + x^4 + x^2 + x + 1, x^7 + x^6 + x^5 + x^4 + x^3 + 1, x^4, x^5 + x^4, x^6 + x^4, x^7 + x^6 + x^5 + x^4, x^3 + x + 1, x^4 + x^3 + x^2 + 1, x^5 + x^2 + x + 1, x^6 + x^5 + x^3 + 1, x^7 + x^5 + x^4 + x^3 + x + 1, x^7 + x^6 + x^4 + x^2 + x, x^6 + x^5 + 1, x^7 + x^5 + x + 1, x^7 + x^6 + x^5 + x^4 + x^3 + x^2 + x, x^4 + x^3 + 1, x^5 + x^3 + x + 1, x^6 + x^5 + x^4 + x^3 + x^2 + 1, x^7 + x^2 + x + 1, x^7 + x^4 + x, x^7 + x^5 + x^3 + x^2 + 1, x^7 + x^6 + x^5 + x^3 + x^2, x^5 + x^3 + x^2 + x + 1, x^6 + x^5 + x^4 + 1, x^7 + x^4 + x + 1, x^7 + x^5 + x^3 + x^2 + x, x^7 + x^6 + x^5 + x^3 + 1, x^5, x^6 + x^5, x^7 + x^5, x^7 + x^6 + x^5 + x^4 + x^3 + x + 1, x^4 + x^2 + x, x^5 + x^4 + x^3 + x, x^6 + x^3 + x^2 + x, x^7 + x^6 + x^4 + x, x^6 + x^5 + x^3 + x^2 + 1, x^7 + x^5 + x^4 + x^2 + x + 1, x^7 + x^6 + x, x^6 + x^4 + x^3 + x^2 + 1, x^7 + x^6 + x^5 + x^2 + x + 1, x^5 + x^4 + x, x^6 + x^4 + x^2 + x, x^7 + x^6 + x^5 + x^4 + x^3 + x, x^4 + x^2 + 1, x^5 + x^4 + x^3 + x^2 + x + 1, x^6 + 1, x^7 + x^6 + x + 1, x^6 + x^4 + x^3 + x^2 + x, x^7 + x^6 + x^5 + x, x^5 + x^4 + x^3 + x^2 + 1, x^6 + x^2 + x + 1, x^7 + x^6 + x^3 + 1, x^6, x^7 + x^6, x^6 + x^4 + x^3 + x + 1, x^7 + x^6 + x^5 + x^3 + x^2 + 1, x^5 + x^3 + x^2, x^6 + x^5 + x^4 + x^2, x^7 + x^4 + x^3 + x^2, x^7 + x^5 + x^4 + x^3 + x^2 + x + 1, x^7 + x^6 + x^4 + x^3 + x, x^6 + x^5 + x^4 + x^2 + 1, x^7 + x^4 + x^3 + x^2 + x + 1, x^7 + x^5 + x^4 + x^3 + x, x^7 + x^6 + x^4 + x^2 + 1, x^6 + x^5 + x^2, x^7 + x^5 + x^3 + x^2, x^7 + x^6 + x^5 + x^3 + x^2 + x + 1, x^5 + x^3 + x, x^6 + x^5 + x^4 + x^3 + x^2 + x, x^7 + x, x^7 + x^4 + x^3 + x^2 + 1, x^7 + x^5 + x^4 + x^3 + x^2, x^7 + x^6 + x^4 + x^3 + x^2 + x + 1, x^6 + x^5 + x^4 + x^3 + x, x^7 + x^3 + x^2 + x, x^7 + x^3 + 1, x^7, x^7 + x^4 + x^3 + x + 1, x^7 + x^5 + x^4 + x^2 + x, x^7 + x^6 + 1, x^6 + x^4 + x^3, x^7 + x^6 + x^5 + x^3, x^5 + x + 1, x^6 + x^5 + x^2 + 1, x^7 + x^5 + x^3 + x^2 + x + 1, x^7 + x^6 + x^5 + x^3 + x, x^5 + x^2 + 1, x^6 + x^5 + x^3 + x^2 + x + 1, x^7 + x^5 + x^4 + 1, x^7 + x^6 + x^3, x^6 + x + 1, x^7 + x^6 + x^2 + 1, x^6 + x^4 + x^2, x^7 + x^6 + x^5 + x^4 + x^3 + x^2, x^4 + x^3 + x^2 + x + 1, x^5 + 1, x^6 + x^5 + x + 1, x^7 + x^5 + x^2 + 1, x^7 + x^6 + x^5 + x^4 + x^2, x^2 + x + 1, x^3 + 1, x^4 + x^3 + x + 1, x^5 + x^3 + x^2 + 1, x^6 + x^5 + x^4 + x^2 + x + 1, x^7 + x^4 + x^3 + 1, x^7 + x^5 + x^4, x^7 + x^6 + x^3 + x + 1, x^6 + x^2 + x, x^7 + x^6 + x^3 + x, x^6 + x^2 + 1, x^7 + x^6 + x^3 + x^2 + x + 1, x^6 + x^3 + x, x^7 + x^6 + x^4 + x^3 + x^2 + x, x^6 + x^5 + x^4 + x^3 + 1, x^7 + x^3 + x + 1, x^7 + x^2 + x, x^7 + x^4 + 1, x^7 + x^5 + x^3, x^7 + x^6 + x^5 + x + 1, x^5 + x^4 + x^3 + x^2 + x, x^6 + x, x^7 + x^6 + x^2 + x, x^6 + x^4 + 1, x^7 + x^6 + x^5 + x^4 + x + 1, x^3 + x^2 + x, x^4 + x, x^5 + x^4 + x^2 + x, x^6 + x^4 + x^3 + x, x^7 + x^6 + x^5 + x^3 + x^2 + x, x^5 + x^3 + 1, x^6 + x^5 + x^4 + x^3 + x + 1, x^7 + x^3 + x^2 + 1, x^7 + x^3 + x^2, x^7 + x^3 + x^2 + x + 1, x^7 + x^3 + x, x^7 + x^2 + 1, x^7 + x^4 + x^2, x^7 + x^5 + x^2 + x + 1, x^7 + x^6 + x^5 + x^4 + x, x^3 + x^2 + 1, x^4 + x^2 + x + 1, x^5 + x^4 + x^3 + 1, x^6 + x^3 + x + 1, x^7 + x^6 + x^4 + x^3 + x^2 + 1, x^6 + x^5 + x^4 + x^3 + x^2, x^7 + x^2, x^7 + x^4 + x^2 + x + 1, x^7 + x^5 + x, x^7 + x^6 + x^5 + x^4 + x^3 + x^2 + 1, x^4 + x^3 + x^2, x^5 + x^2, x^6 + x^5 + x^3 + x^2, x^7 + x^5 + x^4 + x^2, x^7 + x^6 + x^2 + x + 1, x^6 + x^4 + x, x^7 + x^6 + x^5 + x^4 + x^2 + x, 1]"}︡{"stdout":"\n"}︡{"done":true}
︠d08c22cb-1167-42f2-9c1e-42e303631439s︠
len(elements)
︡0577f390-1af9-407c-b0c1-a58a06a05b59︡{"stdout":"256\n"}︡{"done":true}
︠1b08fb5b-f72f-44bf-85c0-2b389eac7ed0s︠
# Examine a random element...
p = elements[129]
p
︡09f2a757-2a20-46bf-a49e-62a2e150f39c︡{"stdout":"x^4 + x^2 + x\n"}︡{"done":true}
︠90baffee-5126-4b0f-b687-f01ac8502b5as︠
p2 = K.fetch_int(0x65)
p2
︡90b56617-4839-4f8b-a700-6c236656670f︡{"stdout":"x^6 + x^5 + x^2 + 1\n"}︡{"done":true}
︠fcf68a59-0770-4a4a-8031-bd9a2922eb57s︠
res = p*p2
hex(res.integer_representation())
︡ff0e05e3-0dec-4547-a4b4-e787af4a3ab0︡{"stdout":"'0x4f'\n"}︡{"done":true}
︠3703c032-620d-4112-88c0-f2264f6df298s︠
#  The polynomial represented as an integer...

p.integer_representation()
︡fae8a71e-91f9-402b-8123-d900c46278ea︡{"stdout":"22\n"}︡{"done":true}
︠45b94b59-f830-4e0a-a66a-3f76e633cf97i︠
%md

Remember the polynomial

$$b_7x^7 + b_6x^6 + b_5x^5 + b_4x^4+b_3x^3 + b_2x^2+b_1x_1 + b_0$$

corresponds to the binary number

$$b_7b_6b_5b_4b_3b_2b_1b_0$$
︡20094867-6d90-403c-a76e-dc9de60db80f︡{"done":true,"md":"\nRemember the polynomial\n\n$$b_7x^7 + b_6x^6 + b_5x^5 + b_4x^4+b_3x^3 + b_2x^2+b_1x_1 + b_0$$\n\ncorresponds to the binary number\n\n$$b_7b_6b_5b_4b_3b_2b_1b_0$$"}
︠d3d6e1ca-3493-4b7b-9b38-03fa7bb97c2as︠
hex(p.integer_representation())
︡82173000-23dd-43af-9c3d-5e6f35968778︡{"stdout":"'0x16'\n"}︡{"done":true}
︠978370ac-9641-4c2d-ac67-0046e9c1f4a9s︠
### The backwards process... Going from a hex number to a polynomial
︡33f7a0e1-e612-463b-b7b2-2810d0b12350︡{"done":true}
︠477b9465-2f6a-4092-8858-7300ce7418efs︠
K.fetch_int(0x17)
︡25fc1d40-ae59-4520-894b-d05d16cf1105︡{"stdout":"x^4 + x^2 + x + 1\n"}︡{"done":true}
︠9e5e12b9-0cbb-4401-8e13-7d2265192c93i︠
%md
### Inverses Table

We now build the "inverses table" for GF[$2^8$].
The one in the book seems to be wrong...

︡87b61ffb-2143-419a-a4cf-2dbdfec6ac34︡{"done":true,"md":"### Inverses Table\n\nWe now build the \"inverses table\" for GF[$2^8$].\nThe one in the book seems to be wrong..."}
︠c9de6d73-b353-4e73-8280-c5e977e9bd44s︠
for row in range(16):
    cur_row = []
    for col in range(16):
        integer = row*16+col
        p = K.fetch_int(integer)
        if not p.is_zero():
            pinv = 1/p
            cur_row.append("0x{:02x}".format(pinv.integer_representation()))
        else:
            cur_row.append("0x00")
    print(cur_row)
︡13f87dd0-b14d-46d0-bcbc-09ffc48188a7︡{"stdout":"['0x00', '0x01', '0x8d', '0xf6', '0xcb', '0x52', '0x7b', '0xd1', '0xe8', '0x4f', '0x29', '0xc0', '0xb0', '0xe1', '0xe5', '0xc7']\n['0x74', '0xb4', '0xaa', '0x4b', '0x99', '0x2b', '0x60', '0x5f', '0x58', '0x3f', '0xfd', '0xcc', '0xff', '0x40', '0xee', '0xb2']\n['0x3a', '0x6e', '0x5a', '0xf1', '0x55', '0x4d', '0xa8', '0xc9', '0xc1', '0x0a', '0x98', '0x15', '0x30', '0x44', '0xa2', '0xc2']\n['0x2c', '0x45', '0x92', '0x6c', '0xf3', '0x39', '0x66', '0x42', '0xf2', '0x35', '0x20', '0x6f', '0x77', '0xbb', '0x59', '0x19']\n['0x1d', '0xfe', '0x37', '0x67', '0x2d', '0x31', '0xf5', '0x69', '0xa7', '0x64', '0xab', '0x13', '0x54', '0x25', '0xe9', '0x09']\n['0xed', '0x5c', '0x05', '0xca', '0x4c', '0x24', '0x87', '0xbf', '0x18', '0x3e', '0x22', '0xf0', '0x51', '0xec', '0x61', '0x17']\n['0x16', '0x5e', '0xaf', '0xd3', '0x49', '0xa6', '0x36', '0x43', '0xf4', '0x47', '0x91', '0xdf', '0x33', '0x93', '0x21', '0x3b']\n['0x79', '0xb7', '0x97', '0x85', '0x10', '0xb5', '0xba', '0x3c', '0xb6', '0x70', '0xd0', '0x06', '0xa1', '0xfa', '0x81', '0x82']\n['0x83', '0x7e', '0x7f', '0x80', '0x96', '0x73', '0xbe', '0x56', '0x9b', '0x9e', '0x95', '0xd9', '0xf7', '0x02', '0xb9', '0xa4']\n['0xde', '0x6a', '0x32', '0x6d', '0xd8', '0x8a', '0x84', '0x72', '0x2a', '0x14', '0x9f', '0x88', '0xf9', '0xdc', '0x89', '0x9a']\n['0xfb', '0x7c', '0x2e', '0xc3', '0x8f', '0xb8', '0x65', '0x48', '0x26', '0xc8', '0x12', '0x4a', '0xce', '0xe7', '0xd2', '0x62']\n['0x0c', '0xe0', '0x1f', '0xef', '0x11', '0x75', '0x78', '0x71', '0xa5', '0x8e', '0x76', '0x3d', '0xbd', '0xbc', '0x86', '0x57']\n['0x0b', '0x28', '0x2f', '0xa3', '0xda', '0xd4', '0xe4', '0x0f', '0xa9', '0x27', '0x53', '0x04', '0x1b', '0xfc', '0xac', '0xe6']\n['0x7a', '0x07', '0xae', '0x63', '0xc5', '0xdb', '0xe2', '0xea', '0x94', '0x8b', '0xc4', '0xd5', '0x9d', '0xf8', '0x90', '0x6b']\n['0xb1', '0x0d', '0xd6', '0xeb', '0xc6', '0x0e', '0xcf', '0xad', '0x08', '0x4e', '0xd7', '0xe3', '0x5d', '0x50', '0x1e', '0xb3']\n['0x5b', '0x23', '0x38', '0x34', '0x68', '0x46', '0x03', '0x8c', '0xdd', '0x9c', '0x7d', '0xa0', '0xcd', '0x1a', '0x41', '0x1c']\n"}︡{"done":true}
︠80b777ee-af65-4b95-ac32-8a8c46c6cfb2s︠
p = K.fetch_int(0x2F)
p
︡139a9d66-1b4e-4991-a79e-dc74ae8d5d2c︡{"stdout":"x^5 + x^3 + x^2 + x + 1\n"}︡{"done":true}
︠42038116-7bc1-4ac2-b0b6-25ecfbd7dd00s︠
1/p
︡7b56755e-8c52-4114-bfb1-1b4bd4d47c54︡{"stdout":"x^6 + x^5 + x^3 + x^2 + x\n"}︡{"done":true}
︠324568b1-dc3b-4e0b-9a8a-4807f5051df3s︠
hex((1/p).integer_representation())
︡4d55ea99-9441-4d0f-a4cb-7768d19c72f3︡{"stdout":"'0x6e'\n"}︡{"done":true}
︠e6747c34-5080-491e-a68b-5dabe9e34e55s︠
p*1/p
︡6092aa2d-30ce-46fd-95a4-9e4702e94b11︡{"stdout":"1\n"}︡{"done":true}
︠266caf4e-a68c-43fd-bd03-58155ce36856s︠
K.modulus()
︡803c214c-3e9a-4291-bab6-20e82df02f7e︡{"stdout":"x^8 + x^4 + x^3 + x^2 + 1\n"}︡{"done":true}
︠30adbf95-6336-4256-9282-a9ba69d3c4b2s︠
q = K.fetch_int(0xC2)
q
︡918ce1ec-33f9-44a8-aec6-d0e4a260ae08︡{"stdout":"x^7 + x^6 + x\n"}︡{"done":true}
︠eaf85ce7-5b2b-405c-90a2-3e3cc370ac6as︠
1/q,hex((1/q).integer_representation())
︡323a524c-7924-48d3-8dab-6cc6df671806︡{"stdout":"(x^7 + x^5 + x^2 + 1, '0xa5')\n"}︡{"done":true}
︠9cde4282-ef68-45da-85d2-fb61fb771ed2s︠
p*q
︡b5c8da6e-3d52-4b2f-b938-b0aef41ca451︡{"stdout":"x^6 + x^3 + x^2 + x + 1\n"}︡{"done":true}
︠a799844a-1f28-4aec-ba01-15f2ad30c702i︠
%md

### The Affine Transformation...

The inverse table is not *quite* the substitution rule for the AES s-box.

One problem is that 0 is a fixed point.

Another is that someone might crack the sbox using algebra.

An "affine transform" solves both problems:

![subbytes](https://slideplayer.com/slide/14424126/90/images/7/AES%3A+Walkthrough+a0+a0+a0+invert+in+GF%2828%29+affine+map+a0+a0+a4+a8+a12.jpg)


The "affine map" can be expressed in the language of linear algebra (mod 2).

![affine_map](https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcSi4ZI8thkYKa9Jg2R-XYRf_7TDd0UtnbzT9cMw_vANUWKaulxT)

This shows how one of the 16 bytes is updated.

︡516554e8-fb6b-4fb1-a6e9-f7f2c8d9dbd0︡{"done":true,"md":"\n### The Affine Transformation...\n\nThe inverse table is not *quite* the substitution rule for the AES s-box.\n\nOne problem is that 0 is a fixed point.  \n\nAnother is that someone might crack the sbox using algebra.\n\nAn \"affine transform\" solves both problems:\n\n![subbytes](https://slideplayer.com/slide/14424126/90/images/7/AES%3A+Walkthrough+a0+a0+a0+invert+in+GF%2828%29+affine+map+a0+a0+a4+a8+a12.jpg)\n\n\nThe \"affine map\" can be expressed in the language of linear algebra (mod 2).\n\n![affine_map](https://encrypted-tbn0.gstatic.com/images?q=tbn:ANd9GcSi4ZI8thkYKa9Jg2R-XYRf_7TDd0UtnbzT9cMw_vANUWKaulxT)\n\nThis shows how one of the 16 bytes is updated."}
︠0aeaa3de-88ad-49fb-ab3e-25f054260513i︠
%md

![afft](afft.png)
︡f60b890e-7adf-41d5-84c5-2bab5446f90d︡{"done":true,"md":"\n![afft](afft.png)"}
︠7a8c1756-0098-410c-b22b-fab4b82717dcs︠
r0 = [1,0,0,0,1,1,1,1]
r1 = [1,1,0,0,0,1,1,1]
r2 = [1,1,1,0,0,0,1,1]
r3 = [1,1,1,1,0,0,0,1]
r4 = [1,1,1,1,1,0,0,0]
r5 = [0,1,1,1,1,1,0,0]
r6 = [0,0,1,1,1,1,1,0]
r7 = [0,0,0,1,1,1,1,1]
M = matrix(GF(2),[r0,r1,r2,r3,r4,r5,r6,r7])

offset = vector(GF(2),[1,1,0,0,0,1,1,0])

sampleinput = vector(GF(2),[1,0,0,0,0,0,0,0])

M*sampleinput + offset
︡43975608-e958-4a5f-9314-7827257a0244︡{"stdout":"(0, 0, 1, 1, 1, 1, 1, 0)\n"}︡{"done":true}
︠43712dc3-065c-442f-a973-4bac4239b7afi︠
%md

### Using the affine transform

Let's do a couple of known sbox inputs and outputs to test the affine transform...

Notice that we have to reverse the output *and* reverse the inverse in GF(8) when typing it in, which is very annoying.

We could fix this by changing $M$ from the book version, but I haven't bothered.

︡1fd63e82-6ab7-4eb7-b59a-2e3a719a1317︡{"done":true,"md":"\n### Using the affine transform\n\nLet's do a couple of known sbox inputs and outputs to test the affine transform...\n\nNotice that we have to reverse the output *and* reverse the inverse in GF(8) when typing it in, which is very annoying.\n\nWe could fix this by changing $M$ from the book version, but I haven't bothered."}
︠413b32a9-af14-4ce7-83ad-b74a323243a8s︠
p=K.fetch_int(0x02)  ## 0x02 --> 0x77
bin((1/p).integer_representation())
︡2d632208-0af3-4e57-9a9e-a3654e285be7︡{"stdout":"'0b10001101'\n"}︡{"done":true}
︠59727e54-7d95-422b-80b6-3c23e48c9d37s︠
v = vector(GF(2),[1,0,1,1,0,0,0,1])
M*v + offset
︡06dcb8f9-1b8a-44e3-9c18-4893ca213b45︡{"stdout":"(1, 1, 1, 0, 1, 1, 1, 0)\n"}︡{"done":true}
︠dae2e4da-f806-49c1-bce9-c812e69d7d94s︠


p=K.fetch_int(0xa1)  ## 0xa1 --> 0x32
bin((1/p).integer_representation())
︡a000136d-b5ff-4e2b-900d-8bdcc93781c2︡{"stdout":"'0b1111100'\n"}︡{"done":true}
︠89351b1a-0c95-48fa-8e08-559930deb72ds︠
v = vector(GF(2),[0,0,1,1,1,1,1,0])
M*v + offset

︡aa421c11-6141-4319-8cd3-113afa3b9373︡{"stdout":"(0, 1, 0, 0, 1, 1, 0, 0)\n"}︡{"done":true}
︠704025cf-304e-4089-857b-6fc8fc05abd7s︠
def AESsbox(hexrep):
    p=K.fetch_int(hexrep)
    bininv = list(bin((1/p).integer_representation())[2:])
    bininv.reverse()
    bininv = list(map(int,bininv+['0']*(8-len(bininv))))
    bininv = vector(GF(2),bininv)
    res = list(M*bininv+offset)
    res.reverse()
    res = "0b"+"".join(map(str,res))
    res = int(res,2)
    return hex(res)
    
    
AESsbox(0x36)    
    
︡19c6cd12-4dc6-428c-b02b-2247a48e22bc︡{"stdout":"'0x5'\n"}︡{"done":true}
︠e0ae351a-aff7-4b9d-8eaf-0ef760451a44︠
︡7e9d2a72-b6a2-4fdc-aee1-4ea151ac8b4c︡
︠bba41047-ca6f-465f-a95b-2148e8948922i︠
%md

![table](https://captanu.files.wordpress.com/2015/04/aes_sbox.jpg)
︡261fd48e-b6dd-457c-8f97-ad9f43c84e23︡{"done":true,"md":"\n![table](https://captanu.files.wordpress.com/2015/04/aes_sbox.jpg)"}
︠006c0c54-0dcb-45ef-8b8c-fcaa6a34a061i︠
%md

### Inverting the sbox...

Obviously we can invert the inversion.

Can we invert the affine transform?

Yes!
︡5e5545b6-bf7e-45ad-aa5a-63ddc2c4f362︡{"done":true,"md":"\n### Inverting the sbox...\n\nObviously we can invert the inversion.\n\nCan we invert the affine transform?\n\nYes!"}
︠831ef319-222d-477d-a8c2-6f745bd57e9e︠
## we use the inverse matrix
M.inverse()
︡9e087688-7650-47b0-bd4c-24a46f84cd43︡{"stdout":"[0 0 1 0 0 1 0 1]\n[1 0 0 1 0 0 1 0]\n[0 1 0 0 1 0 0 1]\n[1 0 1 0 0 1 0 0]\n[0 1 0 1 0 0 1 0]\n[0 0 1 0 1 0 0 1]\n[1 0 0 1 0 1 0 0]\n[0 1 0 0 1 0 1 0]\n"}︡{"done":true}
︠72bf97d5-16ef-4b74-863e-90c5a851e684s︠

v = vector(GF(2),[1,1,1,0,0,1,1,1])

w = M*v + offset

w
︡515aad77-fdaf-484e-988b-1769034caeab︡{"stdout":"(1, 0, 1, 0, 1, 0, 0, 1)\n"}︡{"done":true}
︠18788748-9026-4599-9016-347b0b35dcf9s︠
M.inverse()*(w-offset)
︡1ff41d9d-2db7-41b6-ba4c-1824f5823e33︡{"stdout":"(1, 1, 1, 0, 0, 1, 1, 1)\n"}︡{"done":true}
︠7e67bc5a-0052-4b55-a1c8-d2959037fc7d︠









