︠7243ca0b-24d2-4a6c-a19b-106c58a37921i︠
%md

### The Polynomial Ring over $\mathbb{R}$
In abstract algebra a **ring** is a set $R$ together with a "+" operation and a "$\times$" operation.

The main example of a ring is a space of polynomials with coefficients in $\mathbb{R}$.


We all know that these polynomials can be added and multiplied.

For example if $p = x^3+x^2 + 1$ and $q = x^2+x$ then you can add $p$ and $q$:

$$p + q = (x^3+x^2+1) + (x^2+x) = x^3 + 2x^2 +x + 1$$


You can also multiply polynomials and the result is another polynomial:

$$p\cdot q = (x^3+x^2+1)(x^2+x) = x^5+x^4+2x^3+x^2+x.$$

Polynomials also "distribute" the same way numbers do.  If $p,q,$ and $s$ are polynomials, then

$$p(q+s) = pq + ps.$$


We call algebraic structures that support these operations **rings**.

The integers are also a ring.

--

In many ways polynomials are a lot like integers.

* Like integers, every polynomial $p$ has a "negative" $-p$ and $p+(-p) = 0$.
* Like integers, addition and multiplication on polynomials is commutative and associative:

##### (Commutativity)
$pq = qp$

and

##### (Associativity)
$p(qs) = (pq)s$


* Like integers, you can multiply polynomials, but polynomials do not have "reciprocals".

##### (No inverses)

Just as $\frac{1}{2}$, the inverse of 2, is not an integer, there is similarly no polynomial $p^{-1}$ such that $pp^{-1} = 1$.


### Division theorem

Polynomials and integers are so similar that you can prove some of the same theorems about both.

The **division theorem** for example depends only on properties that integers and polynomials have in common.

**Integer division theorem**

For all integers $n$, $m$ there are integers $q$ and $r$ (with $0 \leq r < m$) such that

$$ n = qm + r$$

**[Polynomial division theorem](https://en.wikipedia.org/wiki/Polynomial_remainder_theorem)**

For all polynomials $p$ and $s$ there are polynomials $q$ and $r$ (with $0 \leq deg($r$) < deg($s$)$) such that

$$ p = qs + r $$


---

#### [Long division](https://www.purplemath.com/modules/polydiv2.htm)

The "proof" of the division theorem in the case of either integers or polynomials is what we call long division.

Long division is basically an algorithm for computing $q$ and $r$.




︡491dc3ca-c600-45a1-b5ad-c018007e5d6b︡{"done":true,"md":"\n### The Polynomial Ring over $\\mathbb{R}$\nIn abstract algebra a **ring** is a set $R$ together with a \"+\" operation and a \"$\\times$\" operation.\n\nThe main example of a ring is a space of polynomials with coefficients in $\\mathbb{R}$.\n\n\nWe all know that these polynomials can be added and multiplied.\n\nFor example if $p = x^3+x^2 + 1$ and $q = x^2+x$ then you can add $p$ and $q$:\n\n$$p + q = (x^3+x^2+1) + (x^2+x) = x^3 + 2x^2 +x + 1$$\n\n\nYou can also multiply polynomials and the result is another polynomial:\n\n$$p\\cdot q = (x^3+x^2+1)(x^2+x) = x^5+x^4+2x^3+x^2+x.$$\n\nPolynomials also \"distribute\" the same way numbers do.  If $p,q,$ and $s$ are polynomials, then \n\n$$p(q+s) = pq + ps.$$\n\n\nWe call algebraic structures that support these operations **rings**.\n\nThe integers are also a ring.\n\n--\n\nIn many ways polynomials are a lot like integers.  \n\n* Like integers, every polynomial $p$ has a \"negative\" $-p$ and $p+(-p) = 0$.\n* Like integers, addition and multiplication on polynomials is commutative and associative:\n\n##### (Commutativity)\n$pq = qp$  \n\nand\n\n##### (Associativity)\n$p(qs) = (pq)s$  \n\n\n* Like integers, you can multiply polynomials, but polynomials do not have \"reciprocals\".\n\n##### (No inverses)\n\nJust as $\\frac{1}{2}$, the inverse of 2, is not an integer, there is similarly no polynomial $p^{-1}$ such that $pp^{-1} = 1$.\n\n\n### Division theorem\n\nPolynomials and integers are so similar that you can prove some of the same theorems about both.\n\nThe **division theorem** for example depends only on properties that integers and polynomials have in common.\n\n**Integer division theorem**\n\nFor all integers $n$, $m$ there are integers $q$ and $r$ (with $0 \\leq r < m$) such that\n\n$$ n = qm + r$$\n\n**[Polynomial division theorem](https://en.wikipedia.org/wiki/Polynomial_remainder_theorem)**\n\nFor all polynomials $p$ and $s$ there are polynomials $q$ and $r$ (with $0 \\leq deg($r$) < deg($s$)$) such that\n\n$$ p = qs + r $$\n\n\n---\n\n#### [Long division](https://www.purplemath.com/modules/polydiv2.htm)\n\nThe \"proof\" of the division theorem in the case of either integers or polynomials is what we call long division.\n\nLong division is basically an algorithm for computing $q$ and $r$."}
︠02c9ecee-eead-4902-a87a-62920b4744b4︠
︡53e30ebd-fbdd-4a62-8dfc-43aa66a3dc8a︡{"done":true}
︠c9d9cf17-5396-48a5-8f57-ff6a58ac02e0i︠
%md

### Polynomials with other kinds of coefficients

We saw that polynomials with coefficients in $\mathbb{R}$ form a ring.

The same thing is true when the coefficients come not from $\mathbb{R}$ but from $\mathbb{Z}_2$.

In this case

$$p + q = (x^3+x^2+1) + (x^2+x) = x^3 + 2x^2 +x + 1 \equiv x^3+x+1.$$

and

$$p\cdot q = (x^3+x^2+1)(x^2+x) = x^5+x^4+2x^3+x^2+x \equiv x^5+x^4+x^2+x.$$


#### Nice properties

You can easily check that we still have all the nice properties as before:

Associativity:  $(pq)s = p(qs)$

Commutativity:  $pq = qp$        (Rings don't actually need to have commutative multiplication.)

Distributivity:  $p(q+s) = pq + ps$.

Negatives:  ??


Actually now each polynomial is it's own negative:

$p + p = 2p \equiv 0$.
︡9e3bfff6-ff97-4bc5-9e8c-8e15875fea5e︡{"done":true,"md":"\n### Polynomials with other kinds of coefficients\n\nWe saw that polynomials with coefficients in $\\mathbb{R}$ form a ring.\n\nThe same thing is true when the coefficients come not from $\\mathbb{R}$ but from $\\mathbb{Z}_2$.\n\nIn this case\n\n$$p + q = (x^3+x^2+1) + (x^2+x) = x^3 + 2x^2 +x + 1 \\equiv x^3+x+1.$$\n\nand\n\n$$p\\cdot q = (x^3+x^2+1)(x^2+x) = x^5+x^4+2x^3+x^2+x \\equiv x^5+x^4+x^2+x.$$\n\n\n#### Nice properties\n\nYou can easily check that we still have all the nice properties as before:\n\nAssociativity:  $(pq)s = p(qs)$\n\nCommutativity:  $pq = qp$        (Rings don't actually need to have commutative multiplication.)\n\nDistributivity:  $p(q+s) = pq + ps$.\n\nNegatives:  ??\n\n\nActually now each polynomial is it's own negative:\n\n$p + p = 2p \\equiv 0$."}
︠e4656eac-40d7-4761-9549-6df521545a03︠
︡7c5697b0-0871-4234-8715-f8da8312559b︡{"done":true}
︠68962b0f-285a-4575-b5ec-3edbb671d826i︠
%md

### Factoring

Everyone in college has some experience factoring polynomials.

For example,  $x^2-4 = (x+2)(x-2)$.

On the other hand you might know that some polynomials "don't factor".

For example $p(x) = x^2 + 1$ has no real roots and no linear factors.

At least it doesn't have them **over the real numbers**.

This means that you can't write $p$ as the product of lower degree polynomials **with real coefficients**.


### $\mathbb{C}$

On the other hand, every college student also knows that over the complex numbers **all** polynomials factor into linear polynomials.


$p(x) = x^2 + 1 = (x+i)(x-i)$, where $i = \sqrt{-1}$.


### Modern algebra

This gives you some important insight:

**Whether a polynomial factors depends on what coefficients you are allowed to use**


### The $\mathbb{Z}_2$ case

Let's use Sage to look at some polynomials with coefficients in $\mathbb{Z}_2$ and see if they factor.

That is:

**Does the polynomial factor into a product of lower degree polynomials *with coefficients in $\mathbb{Z}_2$*?**



︡e0c964ea-6fc2-4869-b72b-25994ac843bc︡{"done":true,"md":"\n### Factoring\n\nEveryone in college has some experience factoring polynomials.\n\nFor example,  $x^2-4 = (x+2)(x-2)$.\n\nOn the other hand you might know that some polynomials \"don't factor\".\n\nFor example $p(x) = x^2 + 1$ has no real roots and no linear factors.\n\nAt least it doesn't have them **over the real numbers**.\n\nThis means that you can't write $p$ as the product of lower degree polynomials **with real coefficients**.\n\n\n### $\\mathbb{C}$\n\nOn the other hand, every college student also knows that over the complex numbers **all** polynomials factor into linear polynomials.\n\n\n$p(x) = x^2 + 1 = (x+i)(x-i)$, where $i = \\sqrt{-1}$.\n\n\n### Modern algebra\n\nThis gives you some important insight:\n\n**Whether a polynomial factors depends on what coefficients you are allowed to use**\n\n\n### The $\\mathbb{Z}_2$ case\n\nLet's use Sage to look at some polynomials with coefficients in $\\mathbb{Z}_2$ and see if they factor.\n\nThat is:\n\n**Does the polynomial factor into a product of lower degree polynomials *with coefficients in $\\mathbb{Z}_2$*?**"}
︠cbd9b3ac-2aef-4b39-a9c6-ddccd916a480︠
R = PolynomialRing(Zmod(2),"x")


︡be9fb05b-c055-4f22-abe5-27382da03094︡{"done":true}
︠ec327372-9d10-4d0a-9b16-a53874d0609d︠
## Here are the elements of degree at most 4
for p in R.polynomials(4):
    print(p)
︡67d545fd-bb83-4561-972a-57b6a4c59949︡{"stdout":"x^4\nx^4 + 1\nx^4 + x\nx^4 + x + 1\nx^4 + x^2\nx^4 + x^2 + 1\nx^4 + x^2 + x\nx^4 + x^2 + x + 1\nx^4 + x^3\nx^4 + x^3 + 1\nx^4 + x^3 + x\nx^4 + x^3 + x + 1\nx^4 + x^3 + x^2\nx^4 + x^3 + x^2 + 1\nx^4 + x^3 + x^2 + x\nx^4 + x^3 + x^2 + x + 1\n"}︡{"done":true}
︠c4ab565b-28db-4c8e-a5dd-fc4ebbf7fed4︠
for p in R.polynomials(4):
    print("{} factors as {}".format(p,p.factor()))
    if p.is_irreducible():
        print("\t\t{} is irreducible!".format(p))
︡613434df-29e2-454f-9c3d-602d3f34fc1d︡{"stdout":"x^4 factors as x^4\nx^4 + 1 factors as (x + 1)^4\nx^4 + x factors as x * (x + 1) * (x^2 + x + 1)\nx^4 + x + 1 factors as x^4 + x + 1\n\t\tx^4 + x + 1 is irreducible!\nx^4 + x^2 factors as x^2 * (x + 1)^2\nx^4 + x^2 + 1 factors as (x^2 + x + 1)^2\nx^4 + x^2 + x factors as x * (x^3 + x + 1)\nx^4 + x^2 + x + 1 factors as (x + 1) * (x^3 + x^2 + 1)\nx^4 + x^3 factors as (x + 1) * x^3\nx^4 + x^3 + 1 factors as x^4 + x^3 + 1\n\t\tx^4 + x^3 + 1 is irreducible!\nx^4 + x^3 + x factors as x * (x^3 + x^2 + 1)\nx^4 + x^3 + x + 1 factors as (x + 1)^2 * (x^2 + x + 1)\nx^4 + x^3 + x^2 factors as x^2 * (x^2 + x + 1)\nx^4 + x^3 + x^2 + 1 factors as (x + 1) * (x^3 + x + 1)\nx^4 + x^3 + x^2 + x factors as x * (x + 1)^3\nx^4 + x^3 + x^2 + x + 1 factors as x^4 + x^3 + x^2 + x + 1\n\t\tx^4 + x^3 + x^2 + x + 1 is irreducible!\n"}︡{"done":true}
︠01b0581b-577d-4bad-a832-7a668edf81e3i︠
%md

### Irreducible polynomials

We say that a polynomial which does not factor into lower degree polynomials (**with coefficients in $\mathbb{Z}_2$**) is **irreducible**.


You can see from the above output that unlike polynomials over $\mathbb{R}$, there can be irreducible polynomials of degree higher than 2.


In fact there are irreducible polynomials of every degree over $\mathbb{Z}_2$.


Irreducible polynomials are a lot like prime numbers.

### Mod out

In fact irreducible polynomials are so much like prime numbers that you can quotient out by them to make interesting algebraic structures.

To see the analogy remember this:

Two **integers** $m$ and $n$ are **equivalent modulo $p\in \mathbb{Z}$** if when you apply the division theorem,

$$ m = q_mp + r_m$$

$$ n = q_np + r_n$$

and $r_m = r_n$  (with $0\leq r_n < q_n$ and $0 \leq r_m < q_m$).

If $p$ is **prime** then the resulting algebraic structure $\mathbb{Z}_p$ is a *field*.

This means that inverses exist for every nonzero element.

(Remember that $a$ has an inverse mod $n$ iff $\gcd(a,n)=1$.  But if $n$ is prime, this is always true if $a \neq 0$.)



︡388bd4db-0efd-4c0c-b3ff-0ba7a5bae01c︡{"done":true,"md":"\n### Irreducible polynomials\n\nWe say that a polynomial which does not factor into lower degree polynomials (**with coefficients in $\\mathbb{Z}_2$**) is **irreducible**.\n\n\nYou can see from the above output that unlike polynomials over $\\mathbb{R}$, there can be irreducible polynomials of degree higher than 2.  \n\n\nIn fact there are irreducible polynomials of every degree over $\\mathbb{Z}_2$.\n\n\nIrreducible polynomials are a lot like prime numbers.\n\n### Mod out\n\nIn fact irreducible polynomials are so much like prime numbers that you can quotient out by them to make interesting algebraic structures.\n\nTo see the analogy remember this:\n\nTwo **integers** $m$ and $n$ are **equivalent modulo $p\\in \\mathbb{Z}$** if when you apply the division theorem,\n\n$$ m = q_mp + r_m$$\n\n$$ n = q_np + r_n$$\n\nand $r_m = r_n$  (with $0\\leq r_n < q_n$ and $0 \\leq r_m < q_m$).\n\nIf $p$ is **prime** then the resulting algebraic structure $\\mathbb{Z}_p$ is a *field*.\n\nThis means that inverses exist for every nonzero element.\n\n(Remember that $a$ has an inverse mod $n$ iff $\\gcd(a,n)=1$.  But if $n$ is prime, this is always true if $a \\neq 0$.)"}
︠49830571-ae42-4681-a53e-0048b0d3f775︠
︡20e2c905-9646-4f35-8bec-7aa2b710fe23︡{"done":true}
︠8278f156-950c-4b01-bf96-37d38a484606i︠
%md

### Polynomial version

Two **polynomials with coefficients in $\mathbb{Z}_2$** $m$ and $n$ are **equivalent modulo $p$ (where $p$ also has coefficients in $\mathbb{Z}_2$)** if when you apply the division theorem,

$$ m = q_mp + r_m$$

$$ n = q_np + r_n$$

and $r_m = r_n$  (with $0\leq \deg(r_n) < \deg(q_n)$ and $0 \leq \deg(r_m) < \deg(q_m)$).

If $p$ is **irreducible** then the resulting algebraic structure is a *field*.

This means that multiplicative inverses exist for every nonzero polynomial.

The form of this algebraic structure depends only on the degree $d$ of $p$.

That is, any other irreducible $p$ of the same degree would produce essentially the same system.

The notation for this algebraic structure is $GF(2^d)$.

This is read "The Galois field of order $2^d$."

There are exactly $2^d$ equivalence classes.

Let's do a Sage example!

︡fb36f454-6b83-4989-840c-ab65486eda6d︡{"done":true,"md":"\n### Polynomial version\n\nTwo **polynomials with coefficients in $\\mathbb{Z}_2$** $m$ and $n$ are **equivalent modulo $p$ (where $p$ also has coefficients in $\\mathbb{Z}_2$)** if when you apply the division theorem,\n\n$$ m = q_mp + r_m$$\n\n$$ n = q_np + r_n$$\n\nand $r_m = r_n$  (with $0\\leq \\deg(r_n) < \\deg(q_n)$ and $0 \\leq \\deg(r_m) < \\deg(q_m)$).\n\nIf $p$ is **irreducible** then the resulting algebraic structure is a *field*.\n\nThis means that multiplicative inverses exist for every nonzero polynomial.\n\nThe form of this algebraic structure depends only on the degree $d$ of $p$. \n\nThat is, any other irreducible $p$ of the same degree would produce essentially the same system.\n\nThe notation for this algebraic structure is $GF(2^d)$.\n\nThis is read \"The Galois field of order $2^d$.\"\n\nThere are exactly $2^d$ equivalence classes.\n\nLet's do a Sage example!"}
︠068d3d06-faa3-4372-bbf3-6576195c9def︠
p
︡f34d6d73-3e0b-4a76-bde5-170330297088︡{"stdout":"x^4 + x^3 + x^2 + x + 1\n"}︡{"done":true}
︠90f8cf47-ed61-4d73-a28a-ca960a82ba15︠
F = GF(2**4,"x",modulus=p)
︡a65057ef-5a99-447a-9b9c-1aacf256ce15︡{"done":true}
︠8fdbcdcd-bc94-4483-922d-f51f10ab30c0︠
for p in F:
    print(p)
︡c7acec1b-f6be-4e80-97fc-babd43a9e633︡{"stdout":"0\nx + 1\nx^2 + 1\nx^3 + x^2 + x + 1\nx^3 + x^2 + x\nx^3 + x^2 + 1\nx^3\nx^2 + x + 1\nx^3 + 1\nx^2\nx^3 + x^2\nx^3 + x + 1\nx\nx^2 + x\nx^3 + x\n1\n"}︡{"done":true}
︠90fc1617-ba7e-4574-a387-ed7e6d711010i︠
%md

### Some observations

Let's make some observations about this field:

1.  There are indeed exactly $2^4 = 16$ elements.
2.  The elements are precisely the polynomials of degree less than $d=4$ with coefficients in $\mathbb{Z}_2$.
3.  These are exactly the possible remainders when you divide a polynomial with coefficients in $\mathbb{Z}_2$ by $p(x) = x^4 + x^3 + x^2 + x + 1$.

Why are there no elements of degree higher than 3?

The reason is that if $s$ has degree 4 or more, then when $p(x) = x^4 + x^3 + x^2 + x + 1$ divides it the remainder has degree at most three.

Therefore every polynomial with coefficients in $\mathbb{Z}_2$ is equivalent,modulo $p$, to some polynomial of degree less than four with coefficients in $\mathbb{Z}_2$.


An interesting exercise is to look at the "powers of $x$".

This is often used in cryptography, for example in the AES key schedule and in XTS-AES block cipher mode.
︡3cf2c5e1-cb46-46ca-85ff-f3a38273e9f7︡{"done":true,"md":"\n### Some observations\n\nLet's make some observations about this field:\n\n1.  There are indeed exactly $2^4 = 16$ elements.\n2.  The elements are precisely the polynomials of degree less than $d=4$ with coefficients in $\\mathbb{Z}_2$.\n3.  These are exactly the possible remainders when you divide a polynomial with coefficients in $\\mathbb{Z}_2$ by $p(x) = x^4 + x^3 + x^2 + x + 1$.\n\nWhy are there no elements of degree higher than 3?\n\nThe reason is that if $s$ has degree 4 or more, then when $p(x) = x^4 + x^3 + x^2 + x + 1$ divides it the remainder has degree at most three.\n\nTherefore every polynomial with coefficients in $\\mathbb{Z}_2$ is equivalent,modulo $p$, to some polynomial of degree less than four with coefficients in $\\mathbb{Z}_2$.\n\n\nAn interesting exercise is to look at the \"powers of $x$\".\n\nThis is often used in cryptography, for example in the AES key schedule and in XTS-AES block cipher mode."}
︠21f62427-41b9-45d7-bcd8-b80973f088e8︠
ex = F("x")
for i in range(20):
    print("x^{:02} = {}".format(i,ex**i))
︡44bd041c-f639-4098-b46b-f03159b4b761︡{"stdout":"x^00 = 1\nx^01 = x\nx^02 = x^2\nx^03 = x^3\nx^04 = x^3 + x^2 + x + 1\nx^05 = 1\nx^06 = x\nx^07 = x^2\nx^08 = x^3\nx^09 = x^3 + x^2 + x + 1\nx^10 = 1\nx^11 = x\nx^12 = x^2\nx^13 = x^3\nx^14 = x^3 + x^2 + x + 1\nx^15 = 1\nx^16 = x\nx^17 = x^2\nx^18 = x^3\nx^19 = x^3 + x^2 + x + 1\n"}︡{"done":true}
︠1ed9eded-8e21-4455-a66a-62e12a738743i︠
%md

### Exercise

Do the reduction $x^4 \equiv x^3 + x^2 + x+ 1 \mod x^4 + x^3 + x^2 + x + 1$ by hand.
︡722d4035-6e2e-4184-97c4-227952e7e824︡{"done":true,"md":"\n### Exercise\n\nDo the reduction $x^4 \\equiv x^3 + x^2 + x+ 1 \\mod x^4 + x^3 + x^2 + x + 1$ by hand."}
︠8caa31a2-3335-4b51-8529-272c6380b008i︠
%md

##### Solution:

Since $p(x) = x^4 + x^3 + x^2 + x + 1 \equiv 0$, it follows that $x^3 + x^2 + x + 1 \equiv x^4$.
︡c94c0b4b-00b9-4590-80f7-b5da64e482de︡{"done":true,"md":"\n##### Solution:\n\nSince $p(x) = x^4 + x^3 + x^2 + x + 1 \\equiv 0$, it follows that $x^3 + x^2 + x + 1 \\equiv x^4$."}
︠f2cc46d6-2f45-41d4-80d7-36893ef77bc5i︠
%md

### Modding out by $x^4+x+1$
The book explores the field $GF(2^4)$ but with the irreducible polynomial $p(x)=x^4+x+1$.

︡fe39ad47-bc2e-47e5-bf3d-695f1d1d9765︡{"done":true,"md":"\n### Modding out by $x^4+x+1$\nThe book explores the field $GF(2^4)$ but with the irreducible polynomial $p(x)=x^4+x+1$."}
︠9a32579e-be10-4ff8-83f2-20540201a98c︠
F = GF(2**4,"x",modulus=x^4+x+1)
︡59bec437-fcd1-4481-8b12-b75adf875429︡{"done":true}
︠935f38e1-e83f-4b2c-8a52-9a5aa0fb59a4︠
for p in F:
    print(p)
︡64e64117-f372-41e4-bca2-cd072b68a7ef︡{"stdout":"0\nx\nx^2\nx^3\nx + 1\nx^2 + x\nx^3 + x^2\nx^3 + x + 1\nx^2 + 1\nx^3 + x\nx^2 + x + 1\nx^3 + x^2 + x\nx^3 + x^2 + x + 1\nx^3 + x^2 + 1\nx^3 + 1\n1\n"}︡{"done":true}
︠2e1bdd1c-6395-4e3f-9a99-b6eee376a091i︠
%md

#### Some similarities and differences

This field is essentially the same a the last one.

But different elements play different "roles".

For instance in this field, powers of $x$ generate all of the nonzero elements.

︡a637a3a6-9f76-4c32-9128-776ce392d645︡{"done":true,"md":"\n#### Some similarities and differences\n\nThis field is essentially the same a the last one.\n\nBut different elements play different \"roles\".\n\nFor instance in this field, powers of $x$ generate all of the nonzero elements."}
︠676aa97f-773a-44d5-a46d-70adf1c8cb9c︠

ex = F("x")
for i in range(16):
    print("x^{:02} = {}".format(i,ex**i))
︡5ddd6a62-df96-4024-80f9-124eef1c3c4a︡{"stdout":"x^00 = 1\nx^01 = x\nx^02 = x^2\nx^03 = x^3\nx^04 = x + 1\nx^05 = x^2 + x\nx^06 = x^3 + x^2\nx^07 = x^3 + x + 1\nx^08 = x^2 + 1\nx^09 = x^3 + x\nx^10 = x^2 + x + 1\nx^11 = x^3 + x^2 + x\nx^12 = x^3 + x^2 + x + 1\nx^13 = x^3 + x^2 + 1\nx^14 = x^3 + 1\nx^15 = 1\n"}︡{"done":true}
︠5c5f4339-e8b9-4d6a-ba8d-0b743c130ddci︠
%md

### Addition in $GF(16)$

Because addition does not increase the degree, addition in GF(16) is basically the same as in the ring of polynomials with coefficients in $\mathbb{Z}_2$.

For example:

$(x^3+x^2+1) + (x^3+x^2) \equiv x^3 + 1$.

You could be a stickler and say what we really care about is not $x^3 + 1$ but rather the remainder when it is divided by $x^4+x+1$.

However because the degree is 3<4, the remainder is just $x^3+1$ itself.


### Multiplication in $GF(16)$

Because all polynomials must have degree at most 3, multiplication in $GF(16)$ is much different than in the polynomial ring over $\mathbb{Z}_2$.


That's because multiplication *does* increase the degree.

For example in the ring $\mathbb{Z}_2[x]$ of polynomials over $\mathbb{Z}_2$,

$(x^3+x^2+1) \cdot (x^3+x^2) = x^6+x^5+x^5+x^4+x^3+x^2 \equiv x^6 + x^4 + x^3 +x^2$.

However we are not interested directly in $x^6 + x^4 + x^3 +x^2$ but rather its *remainder* when we divide it by $x^4+x+1$.

We can use Sage to compute the remainder:

︡87be9ce3-08fd-4f94-89dc-4d025cdf8cca︡{"done":true,"md":"\n### Addition in $GF(16)$\n\nBecause addition does not increase the degree, addition in GF(16) is basically the same as in the ring of polynomials with coefficients in $\\mathbb{Z}_2$.\n\nFor example:\n\n$(x^3+x^2+1) + (x^3+x^2) \\equiv x^3 + 1$.\n\nYou could be a stickler and say what we really care about is not $x^3 + 1$ but rather the remainder when it is divided by $x^4+x+1$.\n\nHowever because the degree is 3<4, the remainder is just $x^3+1$ itself.\n\n\n### Multiplication in $GF(16)$\n\nBecause all polynomials must have degree at most 3, multiplication in $GF(16)$ is much different than in the polynomial ring over $\\mathbb{Z}_2$.\n\n\nThat's because multiplication *does* increase the degree.\n\nFor example in the ring $\\mathbb{Z}_2[x]$ of polynomials over $\\mathbb{Z}_2$,\n\n$(x^3+x^2+1) \\cdot (x^3+x^2) = x^6+x^5+x^5+x^4+x^3+x^2 \\equiv x^6 + x^4 + x^3 +x^2$.\n\nHowever we are not interested directly in $x^6 + x^4 + x^3 +x^2$ but rather its *remainder* when we divide it by $x^4+x+1$.\n\nWe can use Sage to compute the remainder:"}
︠02a29eb3-353d-42fc-950a-74ad21392180︠
R("x^6+x^4+x^3+x^2")%R("x^4+x+1")
︡0c36081b-a52b-4298-bf79-a47b9cf1f333︡{"stdout":"x + 1\n"}︡{"done":true}
︠786ba5c9-b1ef-4568-b43a-5df8a6380717i︠
%md

Note that if we compute the product of the polynomials in the field, we get exactly this result.

︡52250e73-8a52-4906-a341-9c0943474ace︡{"done":true,"md":"\nNote that if we compute the product of the polynomials in the field, we get exactly this result."}
︠71608f87-ed2a-40b9-9c6a-9bdcc2e6cf32︠
F("x^3+x^2+1")*F("x^3+x^2")
︡2ed117c9-fd3a-454c-9ea8-cfcd6070d58c︡{"stdout":"x + 1\n"}︡{"done":true}
︠37852428-1e80-4204-bf57-60544ddb1401i︠
%md

### Funny Arithmetic over Nybbles

We said that $GF(2^4)$ is exactly the polynomials with coefficients in $\mathbb{Z}_2$ with degree less than 4.

More generally, $GF(2^d)$ is exactly the polynomials with coefficients in $\mathbb{Z}_2$ with degree less than $d$.

We can associate polynomials over $\mathbb{Z}_2$ with binary numbers by only remembering the coefficients.

The rule is:

$c_{d-1}x^{d-1} + c_{d-2}x^{d-2} + \cdots + c_1x + c_0 \mapsto c_{d-1}c_{d-2}\cdots c_1 c_ 0$.

For example

$x^3 + x^2 + 1 \mapsto 1101$

and

$x^3 + x^2 \mapsto 1100$

and

$0 \mapsto 0000$

*etc*.

This means that $GF(16)$ exactly corresponds to nybbles (half bytes) or equivalently to the hexadecimal digits.


Here it is in Sage.

︡a32df4a6-6de0-46b5-a1f0-089ab5b583ed︡{"done":true,"md":"\n### Funny Arithmetic over Nybbles\n\nWe said that $GF(2^4)$ is exactly the polynomials with coefficients in $\\mathbb{Z}_2$ with degree less than 4.\n\nMore generally, $GF(2^d)$ is exactly the polynomials with coefficients in $\\mathbb{Z}_2$ with degree less than $d$.\n\nWe can associate polynomials over $\\mathbb{Z}_2$ with binary numbers by only remembering the coefficients.\n\nThe rule is:\n\n$c_{d-1}x^{d-1} + c_{d-2}x^{d-2} + \\cdots + c_1x + c_0 \\mapsto c_{d-1}c_{d-2}\\cdots c_1 c_ 0$.\n\nFor example \n\n$x^3 + x^2 + 1 \\mapsto 1101$\n\nand\n\n$x^3 + x^2 \\mapsto 1100$\n\nand\n\n$0 \\mapsto 0000$\n\n*etc*.\n\nThis means that $GF(16)$ exactly corresponds to nybbles (half bytes) or equivalently to the hexadecimal digits.\n\n\nHere it is in Sage."}
︠65f1196f-30e9-44e5-8802-036aea42fe4e︠
for q in F:
    print("{:17} -> {:0x}".format(str(q),q.integer_representation()))
︡04903409-2099-4567-914f-10970e4987ee︡{"stdout":"0                 -> 0\nx                 -> 2\nx^2               -> 4\nx^3               -> 8\nx + 1             -> 3\nx^2 + x           -> 6\nx^3 + x^2         -> c\nx^3 + x + 1       -> b\nx^2 + 1           -> 5\nx^3 + x           -> a\nx^2 + x + 1       -> 7\nx^3 + x^2 + x     -> e\nx^3 + x^2 + x + 1 -> f\nx^3 + x^2 + 1     -> d\nx^3 + 1           -> 9\n1                 -> 1\n"}︡{"done":true}
︠fb72a380-dd77-48ac-9d55-2562d172ba02i︠
%md

### Funny arithmetic, cont.

To someone who couldn't "see" the polynomials, but only their corresponding digits, it would look like a funny kind of arithmetic is going on.

For example the fact that

$(x^3+x^2+1) + (x^3+x^2) \equiv x^3 + 1$.

would look like

$d + c = 9$.

And the fact that

$(x^3+x^2+1) \cdot (x^3+x^2) = x^6+x^5+x^5+x^4+x^3+x^2 \equiv x^6 + x^4 + x^3 +x^2 \equiv x+1 (\mod x^4+x+1)$

would look like

$d\cdot c = 3$.

We could in fact make a whole crazy multiplication table that would look a little like this:

︡585052e4-e184-4087-80e2-21213103cf70︡{"done":true,"md":"\n### Funny arithmetic, cont.\n\nTo someone who couldn't \"see\" the polynomials, but only their corresponding digits, it would look like a funny kind of arithmetic is going on.\n\nFor example the fact that\n\n$(x^3+x^2+1) + (x^3+x^2) \\equiv x^3 + 1$.\n\nwould look like \n\n$d + c = 9$.\n\nAnd the fact that \n\n$(x^3+x^2+1) \\cdot (x^3+x^2) = x^6+x^5+x^5+x^4+x^3+x^2 \\equiv x^6 + x^4 + x^3 +x^2 \\equiv x+1 (\\mod x^4+x+1)$\n\nwould look like\n\n$d\\cdot c = 3$.\n\nWe could in fact make a whole crazy multiplication table that would look a little like this:"}
︠c1ff32b7-68dd-4953-8258-ada502078714︠
F.multiplication_table(names = [hex(i) for i in range(16)])
︡065361d7-4131-4d12-9e39-4d5b66e46480︡{"stdout":"  *  0x0 0x1 0x2 0x3 0x4 0x5 0x6 0x7 0x8 0x9 0xa 0xb 0xc 0xd 0xe 0xf\n   +----------------------------------------------------------------\n0x0| 0x0 0x0 0x0 0x0 0x0 0x0 0x0 0x0 0x0 0x0 0x0 0x0 0x0 0x0 0x0 0x0\n0x1| 0x0 0x1 0x2 0x3 0x4 0x5 0x6 0x7 0x8 0x9 0xa 0xb 0xc 0xd 0xe 0xf\n0x2| 0x0 0x2 0x4 0x6 0x8 0xa 0xc 0xe 0x3 0x1 0x7 0x5 0xb 0x9 0xf 0xd\n0x3| 0x0 0x3 0x6 0x5 0xc 0xf 0xa 0x9 0xb 0x8 0xd 0xe 0x7 0x4 0x1 0x2\n0x4| 0x0 0x4 0x8 0xc 0x3 0x7 0xb 0xf 0x6 0x2 0xe 0xa 0x5 0x1 0xd 0x9\n0x5| 0x0 0x5 0xa 0xf 0x7 0x2 0xd 0x8 0xe 0xb 0x4 0x1 0x9 0xc 0x3 0x6\n0x6| 0x0 0x6 0xc 0xa 0xb 0xd 0x7 0x1 0x5 0x3 0x9 0xf 0xe 0x8 0x2 0x4\n0x7| 0x0 0x7 0xe 0x9 0xf 0x8 0x1 0x6 0xd 0xa 0x3 0x4 0x2 0x5 0xc 0xb\n0x8| 0x0 0x8 0x3 0xb 0x6 0xe 0x5 0xd 0xc 0x4 0xf 0x7 0xa 0x2 0x9 0x1\n0x9| 0x0 0x9 0x1 0x8 0x2 0xb 0x3 0xa 0x4 0xd 0x5 0xc 0x6 0xf 0x7 0xe\n0xa| 0x0 0xa 0x7 0xd 0xe 0x4 0x9 0x3 0xf 0x5 0x8 0x2 0x1 0xb 0x6 0xc\n0xb| 0x0 0xb 0x5 0xe 0xa 0x1 0xf 0x4 0x7 0xc 0x2 0x9 0xd 0x6 0x8 0x3\n0xc| 0x0 0xc 0xb 0x7 0x5 0x9 0xe 0x2 0xa 0x6 0x1 0xd 0xf 0x3 0x4 0x8\n0xd| 0x0 0xd 0x9 0x4 0x1 0xc 0x8 0x5 0x2 0xf 0xb 0x6 0x3 0xe 0xa 0x7\n0xe| 0x0 0xe 0xf 0x1 0xd 0x3 0x2 0xc 0x9 0x7 0x6 0x8 0x4 0xa 0xb 0x5\n0xf| 0x0 0xf 0xd 0x2 0x9 0x6 0x4 0xb 0x1 0xe 0xc 0x3 0x8 0x7 0x5 0xa\n\n"}︡{"done":true}
︠2c9feaa3-960c-454d-b97c-d632017aa2e7︠

︡883326dd-260a-417f-9f08-db67829fb059︡
︠6c0d14dd-08a7-427a-af59-ad7101b28bbei︠
%md

### Funny arithmetic on bytes

We could do the same thing with bytes rather than nybbles.

We just need a bigger field.
︡fcf6a234-7186-45db-a542-69d88f890160︡{"done":true,"md":"\n### Funny arithmetic on bytes\n\nWe could do the same thing with bytes rather than nybbles.\n\nWe just need a bigger field."}
︠e01365fd-5064-453c-b23b-c670e7c30229s︠
GF256 = GF(2^8,"x",modulus=x^8+x^4+x^3+x+1)

︡842f5a5b-03c8-4af7-a3d8-745c6ec2c9b6︡{"done":true}
︠a70a2b55-aa96-4cbb-b10a-1537a230b65ai︠
%md

This field also has a multiplication table, but now it is size 256x256, which is too big to print out.

Nevertheless, we have our funny math on bytes.

For example the work below shows that

$0x66 \times 0xaf = 0x8b.$


︡b068dab1-96d6-4402-8bc7-bf8a0cbf8c4a︡{"done":true,"md":"\nThis field also has a multiplication table, but now it is size 256x256, which is too big to print out.\n\nNevertheless, we have our funny math on bytes.\n\nFor example the work below shows that\n\n$0x66 \\times 0xaf = 0x8b.$"}
︠9bf1e616-f5b0-4908-b8b3-2521d7c498f0s︠
p
︡4d0f4a0c-49bf-471c-ab79-2f3216412c96︡{"stderr":"Error in lines 1-1\nTraceback (most recent call last):\n  File \"/cocalc/lib/python2.7/site-packages/smc_sagews/sage_server.py\", line 1234, in execute\n    flags=compile_flags), namespace, locals)\n  File \"\", line 1, in <module>\nNameError: name 'p' is not defined\n"}︡{"done":true}
︠24855f19-99da-4818-be82-85cf00995f27︠
p = GF256.fetch_int(0x66)
q = GF256.fetch_int(0xAF)

result = p*q
result
︡c395cbd4-3e2d-4d5e-8b27-fa7a11b6e7ea︡{"stdout":"x^7 + x^3 + x + 1\n"}︡{"done":true}
︠04ae8e9a-2b20-4328-adbd-a6c847e53774︠
print("{:0x}".format(result.integer_representation()))
︡c4732076-04d2-4dff-af2c-b9e5b73ca387︡{"stdout":"8b\n"}︡{"done":true}
︠e802461f-09aa-460a-ae6a-fd991db8385ci︠
%md

### Wait...why are we using a field?

So far we have not made any use of inverses.

We have been using $GF(2^d)$ just as if it were a finite ring (which it also is).

But it turns out that one of the best uses of $GF(2^d)$ is to use the *inverse* operation as a kind of substitution.

If we call this substitution operation, $S$ then

$$S(p) = p^{-1} \mod x^8 + x^4 +x^3+x_1.$$

This turns out to be highly nonlinear.

In cryptography we love non-linear substitutions because they defeat differential cryptanalysis.

Being non-linear in this case means that

$S(p\cdot q) \neq S(p)\cdot S(q).$

We will see later that we also need to tweak this idea a bit to prevent algebraic attacks.

But using inverses for substitutions is an important idea for the AES S-box or substitution box.

We will come back to this when we visit AES in detail.

For now, we use Sage to produce a table of inverses.



︡2e641c6b-08e8-46ee-9ad3-3be4897915cd︡{"done":true,"md":"\n### Wait...why are we using a field?\n\nSo far we have not made any use of inverses.  \n\nWe have been using $GF(2^d)$ just as if it were a finite ring (which it also is).\n\nBut it turns out that one of the best uses of $GF(2^d)$ is to use the *inverse* operation as a kind of substitution.\n\nIf we call this substitution operation, $S$ then\n\n$$S(p) = p^{-1} \\mod x^8 + x^4 +x^3+x_1.$$\n\nThis turns out to be highly nonlinear.\n\nIn cryptography we love non-linear substitutions because they defeat differential cryptanalysis.\n\nBeing non-linear in this case means that \n\n$S(p\\cdot q) \\neq S(p)\\cdot S(q).$\n\nWe will see later that we also need to tweak this idea a bit to prevent algebraic attacks.\n\nBut using inverses for substitutions is an important idea for the AES S-box or substitution box.\n\nWe will come back to this when we visit AES in detail.\n\nFor now, we use Sage to produce a table of inverses."}
︠f67a853f-a7e0-4110-b23d-07c993ae1e8e︠
firstrow="{:8}".format("")
for i in range(16):
    firstrow += "{:6}".format(hex(i))
print(firstrow)
print("-"*len(firstrow))
for i in range(16):
    rowstring = "{:4}:".format(hex(i))
    for j in range(16):
        hexbyte = 16*i+j
        if hexbyte==0:
            rowstring += "{:6}".format(0)
            continue
        p = GF256.fetch_int(hexbyte)
        pinv = 1/p
        rowstring += "{:6x}".format(pinv.integer_representation())
    print(rowstring)
︡aeb7ec90-2c25-4e81-b66f-345e8c023e06︡{"stdout":"        0x0   0x1   0x2   0x3   0x4   0x5   0x6   0x7   0x8   0x9   0xa   0xb   0xc   0xd   0xe   0xf   \n"}︡{"stdout":"--------------------------------------------------------------------------------------------------------\n"}︡{"stdout":"0x0 :     0     1    8d    f6    cb    52    7b    d1    e8    4f    29    c0    b0    e1    e5    c7\n0x1 :    74    b4    aa    4b    99    2b    60    5f    58    3f    fd    cc    ff    40    ee    b2\n0x2 :    3a    6e    5a    f1    55    4d    a8    c9    c1     a    98    15    30    44    a2    c2\n0x3 :    2c    45    92    6c    f3    39    66    42    f2    35    20    6f    77    bb    59    19\n0x4 :    1d    fe    37    67    2d    31    f5    69    a7    64    ab    13    54    25    e9     9\n0x5 :    ed    5c     5    ca    4c    24    87    bf    18    3e    22    f0    51    ec    61    17\n0x6 :    16    5e    af    d3    49    a6    36    43    f4    47    91    df    33    93    21    3b\n0x7 :    79    b7    97    85    10    b5    ba    3c    b6    70    d0     6    a1    fa    81    82\n0x8 :    83    7e    7f    80    96    73    be    56    9b    9e    95    d9    f7     2    b9    a4\n0x9 :    de    6a    32    6d    d8    8a    84    72    2a    14    9f    88    f9    dc    89    9a\n0xa :    fb    7c    2e    c3    8f    b8    65    48    26    c8    12    4a    ce    e7    d2    62\n0xb :     c    e0    1f    ef    11    75    78    71    a5    8e    76    3d    bd    bc    86    57\n0xc :     b    28    2f    a3    da    d4    e4     f    a9    27    53     4    1b    fc    ac    e6\n0xd :    7a     7    ae    63    c5    db    e2    ea    94    8b    c4    d5    9d    f8    90    6b\n0xe :    b1     d    d6    eb    c6     e    cf    ad     8    4e    d7    e3    5d    50    1e    b3\n0xf :    5b    23    38    34    68    46     3    8c    dd    9c    7d    a0    cd    1a    41    1c\n"}︡{"done":true}
︠1a9d8fd6-a20d-4889-96e1-225f44fd7805i︠
%md

### Other finite fields

So far we have explored only finite fields of the form $GF(2^d)$.

These are the finite fields most important to cryptgraphy.

However we are close to a beautiful theorem, so we will state it for your general edification:


---

**Theorem:**

For every prime $p$ and every counting number $d$ there is a unique (up to isomorphism) field of cardinality $p^d$ (denoted $GF(p^d)$). There are no other finite fields.

---

You know what $GF(2^d)$ is:  polynomials over $\mathbb{Z}_2$ modulo some polynomial $P$ which is irreducible over $\mathbb{Z}_2$ and $\deg(P)=d$..

More generally, for any prime $p$,

$GF(p^d)$ is:  polynomials over $\mathbb{Z}_p$ modulo some polynomial $P$ which is irreducible over $\mathbb{Z}_p$ and $\deg(P)=d$.

---

This idea makes sense even when $d=1$, but we usually write $\mathbb{Z}_p$ rather than $GF(p)$ (at least in this class).
︡0a54206e-fe27-4597-acf5-10426f1ca595︡{"done":true,"md":"\n### Other finite fields\n\nSo far we have explored only finite fields of the form $GF(2^d)$.\n\nThese are the finite fields most important to cryptgraphy.\n\nHowever we are close to a beautiful theorem, so we will state it for your general edification:\n\n\n---\n\n**Theorem:**\n\nFor every prime $p$ and every counting number $d$ there is a unique (up to isomorphism) field of cardinality $p^d$ (denoted $GF(p^d)$). There are no other finite fields.\n\n---\n\nYou know what $GF(2^d)$ is:  polynomials over $\\mathbb{Z}_2$ modulo some polynomial $P$ which is irreducible over $\\mathbb{Z}_2$ and $\\deg(P)=d$..\n\nMore generally, for any prime $p$,\n\n$GF(p^d)$ is:  polynomials over $\\mathbb{Z}_p$ modulo some polynomial $P$ which is irreducible over $\\mathbb{Z}_p$ and $\\deg(P)=d$.\n\n---\n\nThis idea makes sense even when $d=1$, but we usually write $\\mathbb{Z}_p$ rather than $GF(p)$ (at least in this class)."}
︠d8e5630d-b080-417c-8dec-20b12ce156c4︠












